import fsspec
import xarray as xr

from .. import setup


def mtbs(store='gcs', return_type='xarray', coarsen=None):
    path = setup.loading(store)
    mapper = fsspec.get_mapper((path / 'processed/mtbs/conus/4000m/raster.zarr').as_uri())
    mtbs = xr.open_zarr(mapper)
    mtbs['x'] = range(len(mtbs['x']))
    mtbs['y'] = range(len(mtbs['y']))
    mtbs = mtbs.drop('x')
    mtbs = mtbs.drop('y')

    if coarsen:
        mtbs = mtbs.coarsen(x=coarsen, y=coarsen, boundary='trim').mean()

    if return_type == 'xarray':
        mtbs.load()
        return mtbs

    if return_type == 'numpy':
        time = mtbs['time'].values
        y = {'burned_area': mtbs['burned_area'].values}

        return y, time



import altair as alt


def plotting():
    alt.renderers.enable('default', embed_options={'actions': False})
    alt.data_transformers.enable('data_server')
    alt.data_transformers.disable_max_rows()



import altair as alt
import pandas as pd


def line(data=None, x=None, y=None, color=None, cmap=None, clim=None, xlim=None, ylim=None):
    """
    plot two variables optionally colored by some feature
    """
    if data is None:
        df = pd.DataFrame({'x': x, 'y': y})
        _x = 'x'
        _y = 'y'
        if color is not None:
            df['color'] = color
            _color = 'color'
    else:
        df = data
        _x = x
        _y = y
        _color = color

    def x_scaled(x):
        return alt.X(x) if xlim is None else alt.X(x, scale=alt.Scale(domain=xlim, clamp=True))

    def y_scaled(y):
        return alt.Y(y) if ylim is None else alt.Y(y, scale=alt.Scale(domain=ylim, clamp=True))

    def color_scaled(color):
        if clim is None and cmap is None:
            return alt.Color(color, scale=alt.Scale(scheme='viridis'))
        elif clim is None and cmap is not None:
            return alt.Color(color, scale=alt.Scale(scheme=cmap))
        elif clim is not None and cmap is None:
            return alt.Color(color, scale=alt.Scale(domain=clim))
        elif clim is not None and cmap is not None:
            return alt.Color(color, scale=alt.Scale(domain=clim, scheme=cmap, clamp=True))

    if color is None:
        line = (
            alt.Chart(df)
            .mark_line(strokeWidth=3, color='rgb(250,100,150)')
            .encode(
                x=x_scaled(_x),
                y=y_scaled(_y),
            )
            .properties(width=350, height=300)
        )
    else:
        line = (
            alt.Chart(df)
            .mark_line(strokeWidth=3)
            .encode(x=x_scaled(_x), y=y_scaled(_y), color=color_scaled(_color))
            .properties(width=350, height=300)
        )

    return line



import os
from collections import defaultdict

import intake
import xarray as xr
import zarr


def make_model_dict(hist_subset, ssp_subset):
    d = defaultdict(list)

    for key_hist in hist_subset:
        left, right = key_hist.rsplit('historical')
        left_scen = left.replace('CMIP', 'ScenarioMIP')
        if not hist_subset[key_hist].df.nunique()['variable_id'] >= 3:
            continue
        for key_ssp in ssp_subset:
            if (
                left_scen in key_ssp
                and right in key_ssp
                and ssp_subset[key_ssp].df.nunique()['variable_id'] >= 3
            ):
                d[key_hist].append(key_ssp)
    model_dict = {k: list(set(v)) for k, v in d.items()}

    return model_dict


def fix_lons(ds):
    lon = ds.lon.copy()
    lon.values[lon.values > 180] -= 360
    ds['lon'] = lon
    return ds


def subset_conus(ds):
    ds = ds.sel(lon=slice(227, 299), lat=slice(19, 55))
    return ds


def subset_time_training(ds):
    return ds.sel(time=slice('1960', '2014'))


def rename(ds):
    if 'longitude' in ds:
        ds = ds.rename({'longitude': 'lon', 'latitude': 'lat'})
    return ds


def preprocess_hist(ds):
    # consider using cmip6_preprocessing here
    return ds.pipe(rename).pipe(subset_conus).pipe(fix_lons).sel(time=slice('1960', '2014'))


def preprocess_ssp(ds):
    # consider using cmip6_preprocessing here
    return ds.pipe(rename).pipe(subset_conus).pipe(fix_lons).sel(time=slice('2015', '2100'))


def cmip():

    col_url = 'https://raw.githubusercontent.com/NCAR/intake-esm-datastore/master/catalogs/pangeo-cmip6.json'
    col = intake.open_esm_datastore(col_url)

    # get all possible simulations
    full_subset = col.search(
        activity_id=['CMIP', 'ScenarioMIP'],
        experiment_id=['historical', 'ssp245', 'ssp370', 'ssp585'],
        member_id='r1i1p1f1',  # remove this at some point
        table_id='Amon',
        grid_label='gn',
        variable_id=['pr', 'tasmin', 'tasmax'],
    )

    # get historical simulations
    hist_subset = full_subset.search(
        activity_id=['CMIP'],
        experiment_id=['historical'],
        require_all_on=['variable_id'],
    )

    # get future simulations
    ssp_subset = full_subset.search(
        activity_id=['ScenarioMIP'],
        experiment_id=['ssp245', 'ssp370', 'ssp585'],
        require_all_on=['variable_id'],
    )

    model_dict = make_model_dict(hist_subset, ssp_subset)

    valid_keys = []
    for k, v in model_dict.items():
        valid_keys.extend([k] + v)

    data = {}
    zarr_kwargs = dict(consolidated=True)
    data.update(hist_subset.to_dataset_dict(zarr_kwargs=zarr_kwargs, preprocess=preprocess_hist))
    data.update(ssp_subset.to_dataset_dict(zarr_kwargs=zarr_kwargs, preprocess=preprocess_ssp))

    for k in list(data):
        if k not in valid_keys:
            del data[k]

    return model_dict, data


def sample_data():

    samples = [
        'CMIP.AWI.AWI-CM-1-1-MR.historical.Amon.gn',
        'CMIP.BCC.BCC-CSM2-MR.historical.Amon.gn',
        'CMIP.CSIRO.ACCESS-ESM1-5.historical.Amon.gn',
        'CMIP.MIROC.MIROC6.historical.Amon.gn',
        'CMIP.HAMMOZ-Consortium.MPI-ESM-1-2-HAM.historical.Amon.gn',
        'CMIP.NUIST.NESM3.historical.Amon.gn',
        'CMIP.CSIRO-ARCCSS.ACCESS-CM2.historical.Amon.gn',
        'CMIP.CCCma.CanESM5.historical.Amon.gn',
        'CMIP.CAS.FGOALS-g3.historical.Amon.gn',
        'CMIP.MPI-M.MPI-ESM1-2-LR.historical.Amon.gn',
        'CMIP.MRI.MRI-ESM2-0.historical.Amon.gn',
        'CMIP.CMCC.CMCC-CM2-SR5.historical.Amon.gn',
        'ScenarioMIP.CCCma.CanESM5.ssp370.Amon.gn',
        'ScenarioMIP.MRI.MRI-ESM2-0.ssp245.Amon.gn',
        'ScenarioMIP.MPI-M.MPI-ESM1-2-LR.ssp245.Amon.gn',
        'ScenarioMIP.BCC.BCC-CSM2-MR.ssp245.Amon.gn',
        'ScenarioMIP.CSIRO.ACCESS-ESM1-5.ssp245.Amon.gn',
        'ScenarioMIP.CSIRO-ARCCSS.ACCESS-CM2.ssp245.Amon.gn',
        'ScenarioMIP.CAS.FGOALS-g3.ssp370.Amon.gn',
        'ScenarioMIP.AWI.AWI-CM-1-1-MR.ssp370.Amon.gn',
        'ScenarioMIP.BCC.BCC-CSM2-MR.ssp370.Amon.gn',
        'ScenarioMIP.CSIRO.ACCESS-ESM1-5.ssp370.Amon.gn',
        'ScenarioMIP.MIROC.MIROC6.ssp370.Amon.gn',
        'ScenarioMIP.AWI.AWI-CM-1-1-MR.ssp585.Amon.gn',
        'ScenarioMIP.MRI.MRI-ESM2-0.ssp585.Amon.gn',
        'ScenarioMIP.CSIRO-ARCCSS.ACCESS-CM2.ssp370.Amon.gn',
        'ScenarioMIP.CAS.FGOALS-g3.ssp245.Amon.gn',
        'ScenarioMIP.MIROC.MIROC6.ssp245.Amon.gn',
        'ScenarioMIP.CMCC.CMCC-CM2-SR5.ssp585.Amon.gn',
        'ScenarioMIP.CSIRO-ARCCSS.ACCESS-CM2.ssp585.Amon.gn',
        'ScenarioMIP.MPI-M.MPI-ESM1-2-LR.ssp585.Amon.gn',
        'ScenarioMIP.MIROC.MIROC6.ssp585.Amon.gn',
        'ScenarioMIP.MPI-M.MPI-ESM1-2-LR.ssp370.Amon.gn',
        'ScenarioMIP.NUIST.NESM3.ssp585.Amon.gn',
        'ScenarioMIP.MRI.MRI-ESM2-0.ssp370.Amon.gn',
        'ScenarioMIP.CAS.FGOALS-g3.ssp585.Amon.gn',
        'ScenarioMIP.CCCma.CanESM5.ssp245.Amon.gn',
        'ScenarioMIP.CMCC.CMCC-CM2-SR5.ssp245.Amon.gn',
        'ScenarioMIP.HAMMOZ-Consortium.MPI-ESM-1-2-HAM.ssp370.Amon.gn',
        'ScenarioMIP.CCCma.CanESM5.ssp585.Amon.gn',
        'ScenarioMIP.CSIRO.ACCESS-ESM1-5.ssp585.Amon.gn',
        'ScenarioMIP.CMCC.CMCC-CM2-SR5.ssp370.Amon.gn',
        'ScenarioMIP.NUIST.NESM3.ssp245.Amon.gn',
        'ScenarioMIP.AWI.AWI-CM-1-1-MR.ssp245.Amon.gn',
        'ScenarioMIP.BCC.BCC-CSM2-MR.ssp585.Amon.gn',
    ]

    cmip_point_data = {}
    for key in samples:
        store = zarr.storage.ABSStore(
            'carbonplan-data',
            prefix=f'carbonplan-scratch/downscaling-point-data/{key}',
            account_name='carbonplan',
            account_key=os.environ['BLOB_ACCOUNT_KEY'],
        )

        cmip_point_data[key] = xr.open_zarr(store)

    store = zarr.storage.ABSStore(
        'carbonplan-data',
        prefix='carbonplan-scratch/downscaling-point-data/terraclimate',
        account_name='carbonplan',
        account_key=os.environ['BLOB_ACCOUNT_KEY'],
    )
    obs_points = xr.open_zarr(store)

    return cmip_point_data, obs_points



import numpy as np
from pyproj import Proj, transform
from rasterio import Affine
from rasterio.crs import CRS
from rasterio.transform import xy


def albers_conus_extent():
    return "-2493045.0 177285.0 2342655.0 3310005.0"


def albers_conus_crs():
    return (
        'PROJCS["Albers_Conical_Equal_Area",'
        'GEOGCS["WGS 84",DATUM["WGS_1984",'
        'SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],'
        "TOWGS84[0,0,0,-0,-0,-0,0],"
        'AUTHORITY["EPSG","6326"]],'
        'PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],'
        'UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],'
        'AUTHORITY["EPSG","4326"]],'
        'PROJECTION["Albers_Conic_Equal_Area"],'
        'PARAMETER["standard_parallel_1",29.5],'
        'PARAMETER["standard_parallel_2",45.5],'
        'PARAMETER["latitude_of_center",23],'
        'PARAMETER["longitude_of_center",-96],'
        'PARAMETER["false_easting",0],'
        'PARAMETER["false_northing",0],'
        'UNIT["meters",1]]'
    )


def albers_conus_transform(res=4000):
    return [res, 0.0, -2493045.0, 0.0, -res, 3310005.0]


def albers_ak_extent():
    return "-2232345.0 344805.0 1494735.0 2380125.0"


def albers_ak_crs():
    return (
        'PROJCS["WGS_1984_Albers",'
        'GEOGCS["WGS 84",DATUM["WGS_1984",'
        'SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],'
        'AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],'
        'UNIT["degree",0.0174532925199433],AUTHORITY["EPSG","4326"]],'
        'PROJECTION["Albers_Conic_Equal_Area"],'
        'PARAMETER["standard_parallel_1",55],'
        'PARAMETER["standard_parallel_2",65],'
        'PARAMETER["latitude_of_center",50],'
        'PARAMETER["longitude_of_center",-154],'
        'PARAMETER["false_easting",0],'
        'PARAMETER["false_northing",0],'
        'UNIT["metre",1,AUTHORITY["EPSG","9001"]]]'
    )


def albers_ak_transform(res=4000):
    return [res, 0.0, -2232345.0, 0.0, -res, 2380125.0]


def rowcol_to_latlon(row, col, res=250):
    row = np.asarray(row) if type(row) is list else row
    col = np.asarray(col) if type(col) is list else col
    x, y = xy(Affine(*albers_conus_transform(res)), row, col)
    p1 = Proj(CRS.from_wkt(albers_conus_crs()))
    p2 = Proj(proj='latlong', datum='WGS84')
    lon, lat = transform(p1, p2, x, y)
    return lat, lon



import warnings

import numpy as np
import xarray as xr
from carbonplan.forests import load
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_auc_score
from tqdm import tqdm

warnings.simplefilter('ignore', category=RuntimeWarning)


def zscore_2d(x, mean=None, std=None):
    recomputing = False
    if mean is None or std is None:
        recomputing = True
    if mean is None:
        mean = x.mean(axis=0)
    if std is None:
        std = x.std(axis=0)
    if recomputing:
        return mean, std, (x - mean) / std
    else:
        return (x - mean) / std


coarsen = None
threshold = 0.25
vars = ['ppt', 'tmax', 'tmax', 'tmax', 'tmin', 'tmin', 'tmin']
aggs = ['sum', 'mean', 'min', 'max', 'mean', 'min', 'max']

print('[fire] fitting model')
y, time = load.mtbs(store='local', return_type='numpy', coarsen=coarsen)

t = len(time)
n, m = y['burned_area'].shape[1:]

X = load.terraclim(
    store='local',
    return_type='numpy',
    coarsen=coarsen,
    tlim=(1984, 2018),
    mean=False,
    vars=vars,
    aggs=aggs,
)

mask = load.nlcd(
    store='local', return_type='numpy', coarsen=coarsen, classes=[41, 42, 43, 90], year=2001
)

all_mask = load.nlcd(store='local', return_type='numpy', coarsen=coarsen, classes='all', year=2001)

final_mask = load.nlcd(
    store='local', return_type='numpy', coarsen=coarsen, classes=[41, 42, 43, 90], year=2016
)
final_mask = final_mask * (final_mask > 0.5)

X['forested'] = np.tile(mask, [t, 1, 1])

for key in X.keys():
    X[key][np.tile(all_mask == 0, [t, 1, 1])] = np.NaN

X_s = np.asarray(
    [np.tile(np.nanmean(X[var], axis=0).flatten(), [1, t]).squeeze() for var in X.keys()]
)

keys = list(X.keys())
keys.remove('forested')
X_t = np.asarray(
    [
        np.tile(np.nanmean(X[var].reshape(t, n * m), axis=1), [n * m, 1]).T.flatten().squeeze()
        for var in keys
    ]
).squeeze()
X_st = np.vstack([X_s, X_t]).T

y_b = y['burned_area'].flatten() > threshold

inds = (~np.isnan(X_st.sum(axis=1))) & (~np.isnan(y_b))

train_mean, train_std, zscored = zscore_2d(X_st[inds])
model = LogisticRegression(fit_intercept=True, max_iter=500, solver='lbfgs')
model.fit(zscored, y_b[inds])

y_hat = model.predict_proba(zscored)[:, 1]

print(roc_auc_score(y_b[inds], y_hat))


print('[fire] evaluating predictions')
targets = list(map(lambda x: str(x), np.arange(2020, 2120, 20)))
scenarios = ['ssp245', 'ssp370', 'ssp585']
ds = xr.Dataset()
for scenario in tqdm(scenarios):
    results = []
    for target in targets:
        tlim = (int(target) - 10, int(target) + 9)

        Xhat = load.cmip(
            model='BCC-CSM2-MR',
            scenario=scenario,
            store='local',
            tlim=tlim,
            coarsen=coarsen,
            mean=False,
            return_type='numpy',
            vars=vars,
            aggs=aggs,
        )

        # for key in Xhat.keys():
        #     Xhat[key] = Xhat[key].reshape([1, n, m])

        t = Xhat['tmax_max'].shape[0]

        Xhat['forested'] = np.tile(mask, [t, 1, 1])

        for key in Xhat.keys():
            Xhat[key][np.tile(all_mask == 0, [t, 1, 1])] = np.NaN

        Xhat_s = np.asarray(
            [
                np.tile(np.nanmean(Xhat[var], axis=0).flatten(), [1, t]).squeeze()
                for var in Xhat.keys()
            ]
        )

        keys = list(Xhat.keys())
        keys.remove('forested')
        Xhat_t = np.asarray(
            [
                np.tile(np.nanmean(Xhat[var].reshape(t, n * m), axis=1), [n * m, 1])
                .T.flatten()
                .squeeze()
                for var in keys
            ]
        ).squeeze()

        Xhat_st = np.vstack([Xhat_s, Xhat_t]).T

        inds = ~np.isnan(Xhat_st.sum(axis=1))

        y_proj = model.predict_proba(zscore_2d(Xhat_st[inds], mean=train_mean, std=train_std))[:, 1]

        y_proj_full = np.zeros((t, n, m)).flatten()
        y_proj_full[inds] = y_proj
        y_proj_full = y_proj_full.reshape(t, n, m).squeeze()
        y_proj_mean = final_mask * np.nanmean(y_proj_full, axis=0)
        results.append(xr.DataArray(y_proj_mean))

    da = xr.concat(results, dim=xr.Variable('year', targets))
    ds[scenario] = da

ds.to_zarr('data/fire.zarr')



import rasterio
import xarray as xr

from .. import setup


def load_rio(f):
    src = rasterio.open(f)
    return src.read(1)


def nlcd(store='gcs', classes=[41, 42, 43, 90], year=2001, return_type='xarray', coarsen=None):
    path = setup.loading(store)

    if classes == 'all':
        classes = [11, 12, 21, 22, 23, 24, 31, 41, 42, 43, 51, 52, 71, 72, 73, 74, 81, 82, 90, 95]

    bands = xr.concat(
        [xr.open_rasterio(path / f'processed/nlcd/conus/4000m/{year}_c{c}.tif') for c in classes],
        dim=xr.Variable('band', classes),
    )
    mask = bands.sum('band', keep_attrs=True)

    if coarsen:
        mask = mask.coarsen(x=coarsen, y=coarsen, boundary='trim').mean()

    if return_type == 'xarray':
        mask.load()
        return mask

    if return_type == 'numpy':
        return mask.values



import pathlib

import urlpath


def loading(store=None):
    if store is None:
        raise ValueError('data store not specified')
    if store == 'gcs':
        base = urlpath.URL('gs://carbonplan-data')
    elif store == 'az':
        base = urlpath.URL('https://carbonplan.blob.core.windows.net/carbonplan-data')
    elif store == 'local':
        base = pathlib.Path(pathlib.Path.home() / 'workdir/carbonplan-data')

    return base



import sys

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from carbonplan.forests import utils

args = sys.argv

if len(args) < 1:
    raise ValueError('must specify dataset')
dataset = args[1]

precision = {'biomass': 2, 'fire': 3}

ds = xr.open_zarr(f'data/{dataset}.zarr')

cmip_model = 'BCC-CSM2-MR'
scenarios = ['ssp245', 'ssp370', 'ssp585']
years = ds['year'].values

a = np.concatenate([ds[scenario].values for scenario in scenarios], axis=0)
a[np.isnan(a)] = 0
r, c = np.nonzero(a.max(axis=0))
lat, lon = utils.rowcol_to_latlon(r, c, res=4000)

df = pd.DataFrame()

for s, scenario in enumerate(scenarios):
    for y, year in enumerate(years):
        key = '0' + '_' + str(s) + '_' + str(y)
        a = ds[scenario].sel(year=year).values
        a[np.isnan(a)] = 0
        df[key] = np.round(a[r, c], precision[dataset])

gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(lon, lat))
gdf.to_file(f'data/{dataset}.geojson', driver='GeoJSON')



from pkg_resources import DistributionNotFound, get_distribution

from . import data, fit, load, plot, preprocess, setup  # noqa

try:
    version = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    version = '0.0.0'  # pragma: no cover
__version__ = version



import altair as alt
import pandas as pd


def carto(
    data=None, lon='lon', lat='lat', projection='albersUsa', color=None, cmap=None, clim=None
):
    if data is None:
        df = pd.DataFrame({'lon': lon, 'lat': lat})
        if color is not None:
            df['color'] = color
            _color = 'color'
    else:
        df = data
        _color = color

    def color_scaled(color):
        if clim is None and cmap is None:
            return alt.Color(color)
        elif clim is None and cmap is not None:
            return alt.Color(color, scale=alt.Scale(scheme=cmap))
        elif clim is not None and cmap is None:
            return alt.Color(color, scale=alt.Scale(domain=clim, clamp=True))
        elif clim is not None and cmap is not None:
            return alt.Color(color, scale=alt.Scale(domain=clim, scheme=cmap, clamp=True))

    if color is None:
        geomap = (
            alt.Chart(df)
            .mark_circle(size=5, color='rgb(150,150,150)')
            .encode(longitude='lon:Q', latitude='lat:Q')
            .project(type=projection)
            .properties(width=650, height=400)
        )
    else:
        geomap = (
            alt.Chart(df)
            .mark_circle(size=3)
            .encode(longitude='lon:Q', latitude='lat:Q', color=color_scaled(_color))
            .project(type=projection)
            .properties(width=650, height=400)
        )

    return geomap



import warnings

import fsspec
import xarray as xr
from pyproj import Proj, transform
from rasterio import Affine
from rasterio.transform import rowcol

from .. import setup, utils


def terraclim(
    store='gcs',
    df=None,
    tlim=None,
    mean=True,
    coarsen=None,
    vars=['ppt', 'tmax'],
    aggs=['sum', 'mean'],
    return_type='xarray',
):

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=ResourceWarning)
        warnings.simplefilter('ignore', category=FutureWarning)
        warnings.simplefilter('ignore', category=RuntimeWarning)

        if tlim is None:
            raise ValueError('must specify time limits as tlim')
        tlim = list(map(str, tlim))

        path = setup.loading(store)
        mapper = fsspec.get_mapper(
            (path / 'processed/terraclimate/conus/4000m/raster.zarr').as_uri()
        )

        ds = xr.open_zarr(mapper)

        X = xr.Dataset()

        def weighted_mean(ds, *args, **kwargs):
            weights = ds.time.dt.days_in_month
            return ds.weighted(weights).mean(dim='time')

        keys = [var + '_' + agg for var, agg in zip(vars, aggs)]

        for key in keys:
            var, agg = key.split('_')
            base = ds[var].resample(time='AS')
            if agg == 'sum':
                X[key] = base.sum('time').sel(time=slice(*tlim))
            elif agg == 'mean':
                X[key] = base.map(weighted_mean, dim='time').sel(time=slice(*tlim))
            elif agg == 'max':
                X[key] = base.max('time').sel(time=slice(*tlim))
            elif agg == 'min':
                X[key] = base.min('time').sel(time=slice(*tlim))
            else:
                raise ValueError(f'agg method {agg} not supported')
            if mean is True:
                X[key] = X[key].mean('time')

        if coarsen:
            X_coarse = xr.Dataset()
            for key in keys:
                X_coarse[key] = X[key].coarsen(x=coarsen, y=coarsen, boundary='trim').mean()
            X = X_coarse

        if df is not None:
            t = Affine(*utils.albers_conus_transform(4000))
            p1 = Proj(utils.albers_conus_crs())
            p2 = Proj(proj='latlong', datum='WGS84')
            x, y = transform(p2, p1, df['lon'].values, df['lat'].values)
            rc = rowcol(t, x, y)
            ind_r = xr.DataArray(rc[0], dims=['x'])
            ind_c = xr.DataArray(rc[1], dims=['x'])
            for key in keys:
                df[key] = X[key][ind_r, ind_c].values
            df = df.dropna().reset_index(drop=True)
            return df

        else:
            if return_type == 'xarray':
                X.load()
                return X

            if return_type == 'numpy':
                X = {key: X[key].values for key in X}
                return X



import pickle

import numpy as np
import pandas as pd
from carbonplan.forests import load  # , fit
from tqdm import tqdm

# import xarray as xr


print('[biomass] loading data')
df = load.fia(store='local', states='conus')
df = load.terraclim(
    store='local', tlim=(2000, 2020), mean=True, df=df, vars=['tmax', 'ppt'], aggs=['mean', 'sum']
)
type_codes = df['type_code'].unique()

# print('[biomass] fitting models')
# models = {}
# for code in tqdm(type_codes):
#     df_type = df[df['type_code'] == code].reset_index()
#     if len(df_type) > 30:
#         x = df_type['age']
#         y = df_type['biomass']
#         f = [df_type['tmax_mean'], df_type['ppt_sum']]
#         model = fit.biomass(x=x, y=y, f=f, noise='gamma')
#         models[code] = model

# with open('data/biomass.pkl', 'wb') as f:
#     pickle.dump(models, f)

with open('data/biomass.pkl', 'rb') as f:
    models = pickle.load(f)

pf = pd.DataFrame()
pf['lat'] = df['lat']
pf['lon'] = df['lon']
pf['type_code'] = df['type_code']

targets = list(map(lambda x: str(x), np.arange(2020, 2120, 20)))

print('[biomass] evaluating predictions')
cmip_model = 'BCC-CSM2-MR'
scenarios = ['ssp245', 'ssp370', 'ssp585']
for it in tqdm(range(len(targets))):
    target = targets[it]
    tlim = (str(int(target) - 10), str(int(target) + 9))
    if target == '2010':
        df = load.terraclim(
            store='local', tlim=tlim, vars=['tmax', 'ppt'], aggs=['mean', 'sum'], mean=True, df=df
        )
    for scenario in scenarios:
        key = cmip_model + '_' + scenario + '_' + target
        if target != '2010':
            df = load.cmip(
                store='local',
                tlim=(int(tlim[0]), int(tlim[1])),
                vars=['tmax', 'ppt'],
                aggs=['mean', 'sum'],
                model=cmip_model,
                scenario=scenario,
                mean=True,
                df=df,
            )
        pf[key] = np.NaN
        for code in type_codes:
            if code in models.keys():
                model = models[code]
                inds = df['type_code'] == code
                x = df[inds]['age']
                year = df[inds]['year']
                f = [df[inds]['tmax_mean'], df[inds]['ppt_sum']]
                if it == 0:
                    pf.loc[inds, key] = model.predict(np.maximum(x + (float(target) - year), 0), f)
                else:
                    prev_target = targets[it - 1]
                    prev_key = cmip_model + '_' + scenario + '_' + prev_target
                    diff = model.predict(x + (float(target) - year), f) - model.predict(
                        x + (float(prev_target) - year), f
                    )
                    pf.loc[inds, key] = np.maximum(pf[prev_key][inds] + diff, 0)

pf = pf.dropna().reset_index(drop=True)
pf['r2'] = pf['type_code'].map(lambda k: models[k].train_r2)
pf['scale'] = pf['type_code'].map(lambda k: models[k].scale)

pf.to_parquet('data/biomass.parquet', compression='gzip', engine='fastparquet')
# pf = pd.read_parquet('data/biomass.parquet')

# print('[biomass] regridding predictions')
# ds = xr.Dataset()
# final_mask = load.nlcd(store='local', year=2016, classes=[41, 42, 43, 90])
# final_mask.values = final_mask.values * (final_mask.values > 0.5)
# for scenario in tqdm(scenarios):
#     results = []
#     for target in targets:
#         key = cmip_model + '_' + scenario + '_' + target
#         gridded = fit.interp(pf, final_mask, var=key)
#         results.append(gridded)
#     da = xr.concat(results, dim=xr.Variable('year', targets))
#     ds[scenario] = da

# ds.to_zarr('data/biomass.zarr')



# flake8: noqa
import math

import networkx as nx
import numpy as np
import pandas as pd


def fia(states, save=True):
    if type(states) is str:
        if states == 'all':
            df = pd.read_csv('gs://carbonplan-data/raw/fia/REF_RESEARCH_STATION.csv')
            return [preprocess_state(state, save=save) for state in df['STATE_ABBR']]
        else:
            return preprocess_state(states, save=save)
    else:
        return [preprocess_state(state, save=save) for state in states]


def tree_based_mortality(tree_df):
    """
    Roll tree-based mortality codes up to condition level.
    TODO: raw agent codes contain per-state info that we might want to use
    """
    tree_df['bulk_agentcd'] = tree_df['AGENTCD'] // 10
    per_agent_mort = (
        tree_df[~np.isnan(tree_df['bulk_agentcd'])]
        .groupby(['PLT_CN', 'CONDID', 'bulk_agentcd'])
        .TPAMORT_UNADJ.sum()
    )
    total_mort = tree_df.groupby(['PLT_CN', 'CONDID']).TPAMORT_UNADJ.sum()

    per_agent_mort = per_agent_mort.reset_index(level=2)

    all_mort = per_agent_mort.join(total_mort, lsuffix='_agent', rsuffix='_total')
    all_mort['fraction'] = all_mort['TPAMORT_UNADJ_agent'] / all_mort['TPAMORT_UNADJ_total']

    bulk_agent_map = {
        1: 'fraction_insect',
        2: 'fraction_disease',
        3: 'fraction_fire',
        8: 'fraction_human',
    }
    fraction_dist = all_mort[['bulk_agentcd', 'fraction']].pivot(columns='bulk_agentcd')['fraction']
    fraction_dist = fraction_dist.rename(columns=bulk_agent_map)[bulk_agent_map.values()]
    return fraction_dist.replace(np.nan, 0).round(2)


def preprocess_state(state_abbr, save=True):
    print(f'preprocessing state {state_abbr}')
    tree_df = pd.read_parquet(
        f'gs://carbonplan-data/raw/fia-states/tree_{state_abbr}.parquet',
        columns=[
            'CN',
            'PLT_CN',
            'DIA',
            'STATUSCD',
            'CONDID',
            'TPA_UNADJ',
            'AGENTCD',
            'TPAMORT_UNADJ',
            'DIACHECK',
            'CARBON_AG',
            'CARBON_BG',
        ],
    )
    # calculate tree-level statistics that will sum later.
    tree_df['unadj_basal_area'] = math.pi * (tree_df['DIA'] / (2 * 12)) ** 2 * tree_df['TPA_UNADJ']

    # 892.179 converts lbs/acre to t/ha
    tree_df['unadj_ag_biomass'] = (
        tree_df['CARBON_AG'] * tree_df['TPA_UNADJ'] * 2 / 892.1791216197013
    )
    tree_df['unadj_bg_biomass'] = (
        tree_df['CARBON_BG'] * tree_df['TPA_UNADJ'] * 2 / 892.1791216197013
    )

    plot_df = pd.read_parquet(f'gs://carbonplan-data/raw/fia-states/plot_{state_abbr}.parquet')
    cond_df = pd.read_parquet(f'gs://carbonplan-data/raw/fia-states/cond_{state_abbr}.parquet')

    cond_vars = [
        'STDAGE',
        'BALIVE',
        'SICOND',
        'SISP',
        'OWNCD',
        'FORTYPCD',
        'FLDTYPCD',
        'DSTRBCD1',
        'DSTRBCD2',
        'DSTRBCD3',
        'TRTCD1',
        'CONDPROP_UNADJ',
        'COND_STATUS_CD',
        'SLOPE',
        'ASPECT',
        'INVYR',
        'CN',
    ]

    cond_agg = cond_df.groupby(['PLT_CN', 'CONDID'])[cond_vars].max()
    cond_agg = cond_agg.join(plot_df.set_index('CN')[['LAT', 'LON', 'ELEV']], on='PLT_CN')

    # TODO: one-hot-encode DSTRBCDs
    def dstrbcd_to_disturb_class(dstrbcd):
        """
        Transforms dstrbcd (int 0-90) to bulk disturbance class (bugs, fires, weather, etc)
        """
        disturb_class_map = {
            1: 'bugs',
            2: 'disease',
            3: 'fire',
            4: 'animal',
            5: 'weather',
            8: 'human',
        }
        return (dstrbcd // 10).map(disturb_class_map)

    hot_encodings = [
        pd.get_dummies(dstrbcd_to_disturb_class(cond_agg[k]), prefix='disturb')
        for k in ['DSTRBCD1', 'DSTRBCD2', 'DSTRBCD3']
    ]
    # sum all disturb codes, then cast to bool so we know if 0/1 disturbance type occurred
    # https://stackoverflow.com/questions/13078751/combine-duplicated-columns-within-a-dataframe
    disturb_flags = (pd.concat(hot_encodings, axis=1)).groupby(level=0, axis=1).sum().astype(bool)

    # per-tree variables that need to sum per condition
    alive_vars = ['unadj_ag_biomass', 'unadj_bg_biomass', 'unadj_basal_area']
    condition_alive_stats = (
        tree_df.loc[tree_df['STATUSCD'] == 1].groupby(['PLT_CN', 'CONDID'])[alive_vars].sum()
    )

    condition_alive_stats = condition_alive_stats.rename(columns={'unadj_basal_area': 'balive'})

    condition_mortality = (
        tree_df.loc[tree_df['TPAMORT_UNADJ'] > 0]
        .groupby(['PLT_CN', 'CONDID'])['unadj_basal_area']
        .sum()
    )

    condition_mortality = condition_mortality.to_frame().rename(
        columns={'unadj_basal_area': 'bamort'}
    )

    tree_mortality_fractions = tree_based_mortality(tree_df)

    full = cond_agg.join(condition_alive_stats)
    full = full.join(disturb_flags)
    full = full.join(condition_mortality)
    full = full.join(tree_mortality_fractions)
    full = full.reset_index()

    full.loc[:, 'adj_mort'] = full.bamort / full.CONDPROP_UNADJ
    full.loc[:, 'adj_balive'] = full.balive / full.CONDPROP_UNADJ
    full.loc[:, 'adj_bg_biomass'] = full.unadj_bg_biomass / full.CONDPROP_UNADJ
    full.loc[:, 'adj_ag_biomass'] = full.unadj_ag_biomass / full.CONDPROP_UNADJ

    plt_uids = generate_uids(plot_df)
    full['plt_uid'] = full['PLT_CN'].map(plt_uids)

    if save:
        full.to_parquet(
            f'gs://carbonplan-data/processed/fia-states/long/{state_abbr}.parquet',
            compression='gzip',
            engine='fastparquet',
        )
    return full


def generate_uids(data, prev_cn_var='PREV_PLT_CN'):
    """
    Generate dict mapping ever CN to a unique group, allows tracking single plot/tree through time
    Can change `prev_cn_var` to apply to tree (etc)
    """
    g = nx.Graph()
    for _, row in data[['CN', prev_cn_var]].iterrows():
        if ~np.isnan(row[prev_cn_var]):  # has ancestor, add nodes + edge
            g.add_edge(row.CN, row[prev_cn_var])
        else:
            g.add_node(row.CN)  # no ancestor, potentially not resampled

    uids = {}
    for uid, subgraph in enumerate(nx.connected_components(g)):
        for node in subgraph:
            uids[node] = uid
    return uids



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("requirements.txt") as f:
    INSTALL_REQUIRES = f.read().strip().split("\n")

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

PYTHON_REQUIRES = ">=3.7"

description = "forest carbon potential and risks"

setup(
    name="carbonplan-forests",
    description=description,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    maintainer="Joe Hamman",
    maintainer_email="joe@carbonplan.org",
    url="https://github.com/carbonplan/forests",
    packages=find_packages(),
    include_package_data=True,
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    tests_require=["pytest"],
    license="MIT",
    keywords="carbon, data, climate",
    use_scm_version={"version_scheme": "post-release", "local_scheme": "dirty-tag"},
    setup_requires=["setuptools_scm", "setuptools>=30.3.0"],
)



import numpy as np
import pyproj
from sklearn.metrics import r2_score

RANDOM_SEED = 1


def score(y_true, y_pred):
    m = ~np.isnan(y_pred)
    if len(y_pred) != m.sum():
        print(f'found {len(m) - m.sum()} nans in y_pred')
    return r2_score(y_true[m], y_pred[m])


def interp(df, mask, var='biomass', spacing=4000):
    """
    Grid a set of lat/lon points to a grid defined by mask

    Parameters
    ----------
    df : pd.DataFrame
        Data points to be gridded in the form of a Pandas DataFrame with
        columns ``lat``, ``lon``, and ``var``.
    mask : xr.DataArray
        Target grid defintion. Must include a pyproj parsable crs attribute
        (e.g. ``mask.attrs['crs']``). Data should be between 0 and 1.
    var : str
        Name of column in df to grid.
    spacing : float
        Grid spacing in units defined by the masks crs.

    Returns
    -------
    grid : xr.DataArray
        Gridded data from df.
    """
    import verde as vd

    # extract the projection and grid info
    region = [mask.x.data[0], mask.x.data[-1], mask.y.data[-1], mask.y.data[0]]
    projection = pyproj.Proj(mask.attrs['crs'])

    coordinates = (df.lon.values, df.lat.values)

    proj_coords = projection(*coordinates)

    # split for validation... this may belong outside of this function
    train, test = vd.train_test_split(
        projection(*coordinates),
        df[var],
        random_state=RANDOM_SEED,
    )

    # fit the gridder
    chain = vd.Chain(
        [
            ('mean', vd.BlockReduce(np.mean, spacing=spacing * 0.25, region=region)),
            ('nearest', vd.ScipyGridder(method='linear')),
        ]
    )

    chain.fit(*train)
    # y_pred = chain.predict(test[0])
    # fit_score = score(test[1][0], y_pred)

    # make the grid
    grid = chain.grid(spacing=spacing, region=region, data_names=[var], dims=('y', 'x'))
    grid = vd.distance_mask(
        proj_coords,
        maxdist=4 * spacing,
        grid=grid,
    )
    grid = np.flipud(grid[var]) * mask
    grid.name = var

    return grid



import altair as alt
import pandas as pd


def xy(
    data=None,
    x=None,
    y=None,
    color=None,
    cmap=None,
    clim=None,
    xlim=None,
    ylim=None,
    xlabel=None,
    ylabel=None,
    clabel=None,
):
    """
    plot two variables optionally colored by some feature
    """
    if data is None:
        if hasattr(x, 'name') and xlabel is None:
            xlabel = x.name
        if hasattr(y, 'name') and ylabel is None:
            ylabel = y.name
        if color is not None and hasattr(color, 'name') and clabel is None:
            clabel = color.name
        df = pd.DataFrame({'x': x, 'y': y})
        _x = 'x'
        _y = 'y'
        if color is not None:
            df['color'] = color
            _color = 'color'
    else:
        if hasattr(data[x], 'name') and xlabel is None:
            xlabel = data[x].name
        if hasattr(data[y], 'name') and ylabel is None:
            ylabel = data[y].name
        if color is not None and hasattr(data[color], 'name') and clabel is None:
            clabel = color.name
        df = data
        _x = x
        _y = y
        _color = color

    if xlabel is None:
        xlabel = 'x'

    if ylabel is None:
        ylabel = 'y'

    if clabel is None:
        clabel = 'color'

    xaxis = alt.Axis(title=xlabel)
    yaxis = alt.Axis(title=ylabel)
    clegend = alt.Legend(title=clabel)

    def x_scaled(x):
        if xlim is None:
            return alt.X(x, axis=xaxis)
        else:
            return alt.X(x, axis=xaxis, scale=alt.Scale(domain=xlim, clamp=True))

    def y_scaled(y):
        if ylim is None:
            alt.Y(y, axis=yaxis)
        else:
            return alt.Y(y, axis=yaxis, scale=alt.Scale(domain=ylim, clamp=True))

    def color_scaled(color):
        if clim is None and cmap is None:
            return alt.Color(color, legend=clegend, scale=alt.Scale(scheme='viridis'))
        elif clim is None and cmap is not None:
            return alt.Color(color, legend=clegend, scale=alt.Scale(scheme=cmap))
        elif clim is not None and cmap is None:
            return alt.Color(color, legend=clegend, scale=alt.Scale(domain=clim))
        elif clim is not None and cmap is not None:
            return alt.Color(
                color, legend=clegend, scale=alt.Scale(domain=clim, scheme=cmap, clamp=True)
            )

    if color is None:
        points = (
            alt.Chart(df)
            .mark_circle(size=42, color='black', fill='black')
            .encode(
                x=x_scaled(_x),
                y=y_scaled(_y),
            )
            .properties(width=350, height=300)
        )
    else:
        points = (
            alt.Chart(df)
            .mark_circle(size=42)
            .encode(x=x_scaled(_x), y=y_scaled(_y), color=color_scaled(_color))
            .properties(width=350, height=300)
        )

    return points



import warnings

import numpy as np
from scipy.optimize import Bounds, minimize
from scipy.stats import gamma, norm


def logistic(x, f, p):
    a, b, c, w0, w1 = p
    return (
        (a + w0 * f[0] + w1 * f[1])
        * (1 / (1 + c * np.exp(-b * (x))) - (1 / (1 + c)))
        * ((c + 1) / c)
    )


def biomass(x, y, f, noise='gamma', init=None):
    def loglik(x, y, f, p):
        a, b, c, w0, w1, scale = p
        _mu = logistic(x, f, [a, b, c, w0, w1])
        if noise == 'gamma':
            return -np.sum(gamma.logpdf(y, np.maximum(_mu / scale, 1e-12), scale=scale))
        if noise == 'normal':
            return -np.sum(norm.logpdf(y, loc=_mu, scale=scale))

    fx = lambda p: loglik(x, y, f, p)

    lb = np.ones((6,)) * 0.00001
    lb[2] = 1
    lb[3] = -np.inf
    lb[4] = -np.inf
    ub = np.ones((6,)) * np.inf
    bounds = Bounds(lb, ub)
    if init is None:
        init = [np.nanmean(y), 0.1, 10, 0, 0, np.nanstd(y)]
    # options_lbfgsb = {'eps': 1e-10, 'maxcor': 100, 'maxiter': 500}
    options_trust = {'maxiter': 5000}

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=UserWarning)
        result = minimize(fx, init, bounds=bounds, method='trust-constr', options=options_trust)

    if result.success is False:
        print('optimization failed')

    return Model(result, noise, x, y, f)


class Model:
    def __init__(self, result, noise, x=None, y=None, f=None):
        self.result = result
        self.noise = noise
        self.p = result.x[0 : len(result.x) - 1]
        self.scale = result.x[len(result.x) - 1]
        if x is not None and y is not None and f is not None:
            self.train_r2 = self.r2(x, f, y)

    def __repr__(self):
        return str(self.result)

    def r2(self, x, f, y):
        yhat = self.predict(x, f)
        return 1 - np.nanvar(y - yhat, ddof=1) / np.nanvar(y, ddof=1)

    def predict(self, x, f, percentile=None):
        if percentile is not None:
            f = [np.nanpercentile(f[0], percentile[0]), np.nanpercentile(f[1], percentile[1])]
        return logistic(x, f, self.p)

    def sample(self, x, f):
        mu = logistic(x, f, self.p)
        if self.noise == 'gamma':
            return np.random.gamma(mu / self.scale, scale=self.scale)
        if self.noise == 'normal':
            return np.random.normal(mu, scale=self.scale)
